use anyhow::Error;
use piccolo::{
    Callback, Closure, Context, Executor, FromValue, Function, IntoValue, Lua, StashedExecutor,
    StaticError, Table,
};

use crate::{echo_to_term_frame, GlobalCell, TermFrame};
use std::str;

pub struct LuaState {
    pub interp: Lua,
    pub exec: StashedExecutor,
}

impl LuaState {
    pub fn setup() -> Result<LuaState, String> {
        let mut interp = Lua::core();
        let exec: StashedExecutor =
            interp.enter(|ctx| Ok::<StashedExecutor, String>(ctx.stash(Executor::new(ctx))))?;

        Ok(LuaState { interp, exec })
    }

    fn try_set_current_frame(&mut self, frame: &TermFrame) -> Result<(), StaticError> {
        self.interp.try_enter(|ctx| {
            let info_table = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"info")))?;
            info_table.set(ctx, ctx.intern_static(b"current_frame"), frame.0 as i64)?;
            Ok(())
        })
    }

    pub fn set_current_frame(&mut self, frame: &TermFrame) {
        // We silently ignore errors here. Failure can happen if the Lua code does weird things
        // like messes with the info global, so better to just ignore.
        self.try_set_current_frame(frame).unwrap_or(())
    }

    pub fn execute(&mut self, command: &str) -> Result<(), String> {
        self.interp
            .try_enter(|ctx| {
                let closure = Closure::load(ctx, None, format!("return ({})", command).as_bytes())
                    .or_else(|_| Closure::load(ctx, None, command.as_bytes()))?;
                ctx.fetch(&self.exec).restart(ctx, closure.into(), ());
                Ok(())
            })
            .map_err(|err| format!("{}", err))?;
        self.interp
            .execute::<()>(&self.exec)
            .map_err(|err| format!("{}", err))
    }
}

pub fn install_lua_globals(global: &GlobalCell) -> Result<(), String> {
    global
        .lua_engine
        .borrow_mut()
        .interp
        .try_enter(|ctx| {
            let cmd_table = Table::new(&ctx);
            macro_rules! register_command {
                ($sym: ident) => {
                    cmd_table
                        .set(
                            ctx,
                            ctx.intern_static(stringify!($sym).as_bytes()),
                            $sym(ctx, global.clone()),
                        )
                        .map_err(|_| Error::msg("Can't add command"))?;
                };
            }

            register_command!(echo_frame_raw);
            register_command!(echo_frame);
            register_command!(echo);
            ctx.set_global(ctx.intern_static(b"commands").into_value(ctx), cmd_table)
                .map(|_| ())
                .map_err(|_| Error::msg("Can't set commands key"))?;

            let info_table = Table::new(&ctx);
            ctx.set_global(ctx.intern_static(b"info").into_value(ctx), info_table)
                .map(|_| ())
                .map_err(|_| Error::msg("Can't set info key"))?;

            Ok(())
        })
        .map_err(|e| e.to_string())?;

    Ok(())
}

fn echo_frame_raw(ctx: Context, global: GlobalCell) -> Callback {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame_no: u64 = stack.from_front(ctx)?;
        let message: piccolo::String = stack.from_front(ctx)?;
        let message_str = str::from_utf8(message.as_bytes())
            .map_err(|_| "Expected message to echo to be UTF-8.".into_value(ctx))?;
        echo_to_term_frame(&global, &TermFrame(frame_no), message_str)
            .map_err(|m| m.into_value(ctx))?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn echo_frame<'gc>(ctx: Context<'gc>, _global: GlobalCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> =
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?;
        let function = Function::from_value(ctx, commands.get(ctx, "echo_frame_raw"))?;
        let message: piccolo::Value = stack.from_back(ctx)?;
        let message = ctx.intern(&format!("{}\r\n", message).as_bytes());
        stack.push_back(message.into());
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

fn echo<'gc>(ctx: Context<'gc>, _global: GlobalCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> =
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?;
        let function = Function::from_value(ctx, commands.get(ctx, "echo_frame"))?;
        let info: Table<'gc> = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"info")))?;
        let cur_frame = info.get(ctx, ctx.intern_static(b"current_frame"));
        stack.push_front(cur_frame);
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}
